package ru.dbotthepony.kstarbound

import com.github.benmanes.caffeine.cache.Cache
import com.github.benmanes.caffeine.cache.Caffeine
import com.google.common.cache.CacheBuilder
import com.google.common.collect.Interner
import com.google.common.collect.Interners
import com.google.gson.*
import com.google.gson.internal.bind.JsonTreeReader
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import org.apache.logging.log4j.LogManager
import org.lwjgl.stb.STBImage
import ru.dbotthepony.kstarbound.api.ISBFileLocator
import ru.dbotthepony.kstarbound.api.IStarboundFile
import ru.dbotthepony.kstarbound.api.NonExistingFile
import ru.dbotthepony.kstarbound.api.PhysicalFile
import ru.dbotthepony.kstarbound.api.explore
import ru.dbotthepony.kstarbound.defs.*
import ru.dbotthepony.kstarbound.defs.image.AtlasConfiguration
import ru.dbotthepony.kstarbound.defs.image.ImageReference
import ru.dbotthepony.kstarbound.defs.item.BackArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.ChestArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.CurrencyItemPrototype
import ru.dbotthepony.kstarbound.defs.item.FlashlightPrototype
import ru.dbotthepony.kstarbound.defs.item.HarvestingToolPrototype
import ru.dbotthepony.kstarbound.defs.item.HeadArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.IArmorItemDefinition
import ru.dbotthepony.kstarbound.defs.item.IItemDefinition
import ru.dbotthepony.kstarbound.defs.item.InventoryIcon
import ru.dbotthepony.kstarbound.defs.item.ItemPrototype
import ru.dbotthepony.kstarbound.defs.item.LegsArmorItemPrototype
import ru.dbotthepony.kstarbound.defs.item.LiquidItemPrototype
import ru.dbotthepony.kstarbound.defs.item.MaterialItemPrototype
import ru.dbotthepony.kstarbound.defs.npc.NpcTypeDefinition
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.defs.particle.ParticleDefinition
import ru.dbotthepony.kstarbound.defs.player.BlueprintLearnList
import ru.dbotthepony.kstarbound.defs.player.PlayerDefinition
import ru.dbotthepony.kstarbound.defs.player.RecipeDefinition
import ru.dbotthepony.kstarbound.defs.player.TechDefinition
import ru.dbotthepony.kstarbound.defs.projectile.ProjectileDefinition
import ru.dbotthepony.kstarbound.defs.quest.QuestTemplate
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.io.*
import ru.dbotthepony.kstarbound.io.json.AABBTypeAdapter
import ru.dbotthepony.kstarbound.io.json.AABBiTypeAdapter
import ru.dbotthepony.kstarbound.io.json.EitherTypeAdapter
import ru.dbotthepony.kstarbound.io.json.InternedJsonElementAdapter
import ru.dbotthepony.kstarbound.io.json.InternedStringAdapter
import ru.dbotthepony.kstarbound.io.json.LongRangeAdapter
import ru.dbotthepony.kstarbound.io.json.NothingAdapter
import ru.dbotthepony.kstarbound.io.json.Vector2dTypeAdapter
import ru.dbotthepony.kstarbound.io.json.Vector2fTypeAdapter
import ru.dbotthepony.kstarbound.io.json.Vector2iTypeAdapter
import ru.dbotthepony.kstarbound.io.json.Vector4dTypeAdapter
import ru.dbotthepony.kstarbound.io.json.Vector4iTypeAdapter
import ru.dbotthepony.kstarbound.io.json.builder.EnumAdapter
import ru.dbotthepony.kstarbound.io.json.builder.BuilderAdapter
import ru.dbotthepony.kstarbound.io.json.builder.FactoryAdapter
import ru.dbotthepony.kstarbound.io.json.builder.JsonImplementationTypeFactory
import ru.dbotthepony.kstarbound.io.json.factory.ArrayListAdapterFactory
import ru.dbotthepony.kstarbound.io.json.factory.ImmutableCollectionAdapterFactory
import ru.dbotthepony.kstarbound.lua.LuaState
import ru.dbotthepony.kstarbound.lua.loadInternalScript
import ru.dbotthepony.kstarbound.math.*
import ru.dbotthepony.kstarbound.util.ItemStack
import ru.dbotthepony.kstarbound.util.PathStack
import ru.dbotthepony.kstarbound.util.SBPattern
import ru.dbotthepony.kstarbound.util.WriteOnce
import ru.dbotthepony.kstarbound.util.traverseJsonPath
import ru.dbotthepony.kvector.vector.nint.Vector2i
import java.io.*
import java.lang.ref.WeakReference
import java.text.DateFormat
import java.time.Duration
import java.util.*
import java.util.concurrent.locks.LockSupport
import java.util.function.BiConsumer
import java.util.function.BinaryOperator
import java.util.function.Function
import java.util.function.Supplier
import java.util.stream.Collector
import kotlin.NoSuchElementException
import kotlin.collections.ArrayList

class Starbound : ISBFileLocator {
	private val logger = LogManager.getLogger()

	val pathStack = PathStack(STRINGS)

	private val _tiles = ObjectRegistry("tiles", TileDefinition::materialName, TileDefinition::materialId)
	val tiles = _tiles.view
	val tilesByID = _tiles.intView

	private val _tileModifiers = ObjectRegistry("tile modifiers", MaterialModifier::modName, MaterialModifier::modId)
	val tileModifiers = _tileModifiers.view
	val tileModifiersByID = _tileModifiers.intView

	private val _liquid = ObjectRegistry("liquid", LiquidDefinition::name, LiquidDefinition::liquidId)
	val liquid = _liquid.view
	val liquidByID = _liquid.intView

	private val _species = ObjectRegistry("species", Species::kind)
	val species = _species.view

	private val _statusEffects = ObjectRegistry("status effects", StatusEffectDefinition::name)
	val statusEffects = _statusEffects.view

	private val _particles = ObjectRegistry("particles", ParticleDefinition::kind)
	val particles = _particles.view

	private val _items = ObjectRegistry("items", IItemDefinition::itemName)
	val items = _items.view

	private val _questTemplates = ObjectRegistry("quest templates", QuestTemplate::id)
	val questTemplates = _questTemplates.view

	private val _techs = ObjectRegistry("techs", TechDefinition::name)
	val techs = _techs.view

	private val _jsonFunctions = ObjectRegistry<JsonFunction>("json functions")
	val jsonFunctions = _jsonFunctions.view

	private val _json2Functions = ObjectRegistry<Json2Function>("json 2functions")
	val json2Functions = _json2Functions.view

	private val _npcTypes = ObjectRegistry("npc types", NpcTypeDefinition::type)
	val npcTypes = _npcTypes.view

	private val _projectiles = ObjectRegistry("projectiles", ProjectileDefinition::projectileName)
	val projectiles = _projectiles.view

	val recipeRegistry = RecipeRegistry()

	val gson: Gson = with(GsonBuilder()) {
		serializeNulls()
		setDateFormat(DateFormat.LONG)
		setFieldNamingPolicy(FieldNamingPolicy.IDENTITY)
		setPrettyPrinting()

		registerTypeAdapter(InternedStringAdapter(STRINGS))

		InternedJsonElementAdapter(STRINGS).also {
			registerTypeAdapter(it)
			registerTypeAdapter(it.arrays)
			registerTypeAdapter(it.objects)
		}

		registerTypeAdapter(Nothing::class.java, NothingAdapter)

		// Обработчик @JsonImplementation
		registerTypeAdapterFactory(JsonImplementationTypeFactory)

		// ImmutableList, ImmutableSet, ImmutableMap
		registerTypeAdapterFactory(ImmutableCollectionAdapterFactory(STRINGS))

		// ArrayList
		registerTypeAdapterFactory(ArrayListAdapterFactory)

		// все enum'ы без особых настроек
		registerTypeAdapterFactory(EnumAdapter.Companion)

		// @JsonBuilder
		registerTypeAdapterFactory(BuilderAdapter.Factory(STRINGS))

		// @JsonFactory
		registerTypeAdapterFactory(FactoryAdapter.Factory(STRINGS))

		// Either<>
		registerTypeAdapterFactory(EitherTypeAdapter)
		registerTypeAdapterFactory(SBPattern.Companion)

		registerTypeAdapter(ColorReplacements.Companion)
		registerTypeAdapterFactory(BlueprintLearnList.Companion)

		registerTypeAdapter(ColorTypeAdapter.nullSafe())

		// математические классы
		registerTypeAdapter(AABBTypeAdapter)
		registerTypeAdapter(AABBiTypeAdapter)
		registerTypeAdapter(Vector2dTypeAdapter)
		registerTypeAdapter(Vector2fTypeAdapter)
		registerTypeAdapter(Vector2iTypeAdapter)
		registerTypeAdapter(Vector4iTypeAdapter)
		registerTypeAdapter(Vector4dTypeAdapter)
		registerTypeAdapter(PolyTypeAdapter)

		// Функции
		registerTypeAdapter(JsonFunction.CONSTRAINT_ADAPTER)
		registerTypeAdapter(JsonFunction.INTERPOLATION_ADAPTER)
		registerTypeAdapter(JsonFunction.Companion)
		registerTypeAdapterFactory(Json2Function.Companion)

		// Общее
		registerTypeAdapterFactory(ThingDescription.Factory(STRINGS))

		registerTypeAdapter(EnumAdapter(DamageType::class, default = DamageType.NORMAL))

		registerTypeAdapterFactory(InventoryIcon.Factory(pathStack))

		registerTypeAdapterFactory(IArmorItemDefinition.Frames.Factory)
		registerTypeAdapterFactory(DirectAssetReferenceFactory(pathStack))
		registerTypeAdapterFactory(ImageReference.Factory({ atlasRegistry.get(it) }, pathStack))

		registerTypeAdapterFactory(AssetReferenceFactory(pathStack, this@Starbound))

		registerTypeAdapter(ItemStack.Adapter(this@Starbound))

		registerTypeAdapterFactory(ItemReference.Factory(STRINGS))

		registerTypeAdapterFactory(with(RegistryReferenceFactory()) {
			add(tiles::get)
			add(tileModifiers::get)
			add(liquid::get)
			add(items::get)
			add(species::get)
			add(statusEffects::get)
			add(particles::get)
		})

		registerTypeAdapter(LongRangeAdapter)

		create()
	}

	val atlasRegistry = AtlasConfiguration.Registry(this, pathStack, gson)

	private val imageCache: Cache<String, ImageData> = Caffeine.newBuilder()
		.softValues()
		.expireAfterAccess(Duration.ofMinutes(20))
		.weigher<String, ImageData> { key, value -> value.data.capacity() }
		.maximumWeight(1_024L * 1_024L * 256L /* 256 МиБ */)
		.build()

	init {
		val ref = WeakReference(imageCache)

		val worker = Runnable {
			while (true) {
				val get = ref.get() ?: break
				get.cleanUp()
				LockSupport.parkNanos(1_000_000_000L)
			}
		}

		Thread(worker, "Image Data Cache Cleaner for $this").also {
			it.isDaemon = true
			it.start()
		}
	}

	fun item(name: String): ItemStack {
		return ItemStack(items[name] ?: return ItemStack.EMPTY)
	}

	fun item(name: String, count: Long): ItemStack {
		if (count <= 0L)
			return ItemStack.EMPTY

		return ItemStack(items[name] ?: return ItemStack.EMPTY, count = count)
	}

	fun item(name: String, count: Long, parameters: JsonObject): ItemStack {
		if (count <= 0L)
			return ItemStack.EMPTY

		return ItemStack(items[name] ?: return ItemStack.EMPTY, count = count, parameters = parameters)
	}

	fun item(descriptor: JsonObject): ItemStack {
		return item(
			(descriptor["name"] as? JsonPrimitive)?.asString ?: return ItemStack.EMPTY,
			descriptor["count"]?.asLong ?: return ItemStack.EMPTY,
			(descriptor["parameters"] as? JsonObject)?.deepCopy() ?: JsonObject()
		)
	}

	fun item(descriptor: JsonElement?): ItemStack {
		if (descriptor is JsonPrimitive) {
			return item(descriptor.asString)
		} else if (descriptor is JsonObject) {
			return item(descriptor)
		} else {
			return ItemStack.EMPTY
		}
	}

	var initializing = false
		private set
	var initialized = false
		private set

	@Volatile
	var terminateLoading = false

	fun loadJsonAsset(path: String): JsonElement? {
		val filename: String
		val jsonPath: String?

		if (path.contains(':')) {
			filename = path.substringBefore(':')
			jsonPath = path.substringAfter(':')
		} else {
			filename = path
			jsonPath = null
		}

		val file = locate(filename)

		if (!file.isFile) {
			return null
		}

		return traverseJsonPath(jsonPath, gson.fromJson(file.reader(), JsonElement::class.java))
	}

	private fun luaRequire(it: LuaState, args: LuaState.ArgStack) {
		val name = args.getString()
		val file = locate(name)

		if (!file.exists) {
			throw FileNotFoundException("File $name does not exist")
		}

		if (!file.isFile) {
			throw FileNotFoundException("File $name is a directory")
		}

		val read = file.readToString()
		it.load(read, chunkName = "@" + file.computeFullPath())
		it.call()
	}

	fun imageData(path: String): ImageData {
		return imageCache.get(path) {
			val file = locate(path)

			if (!file.exists) {
				throw FileNotFoundException("No such file $file")
			}

			if (!file.isFile) {
				throw IllegalStateException("File $file is a directory")
			}

			val getWidth = intArrayOf(0)
			val getHeight = intArrayOf(0)
			val components = intArrayOf(0)

			val data = STBImage.stbi_load_from_memory(
				file.readDirect(),
				getWidth, getHeight,
				components, 0
			) ?: throw IllegalArgumentException("File $file is not an image or it is corrupted")

			ImageData(data, getWidth[0], getHeight[0], components[0])
		}
	}

	fun imageSize(path: String): Vector2i {
		val image = imageData(path)
		return Vector2i(image.width, image.height)
	}

	fun pushLuaAPI(state: LuaState) {
		state.pushWeak(this) { args ->
			luaRequire(args.lua, args)
			0
		}

		state.storeGlobal("require")

		state.pushTable()
		state.storeGlobal("root")
		state.loadGlobal("root")

		state.setTableFunction("assetJson", this) {args ->
			args.lua.push(loadJsonAsset(args.getString()))
			1
		}

		state.setTableFunction("makeCurrentVersionedJson", this) {args ->
			TODO("makeCurrentVersionedJson")
		}

		state.setTableFunction("loadVersionedJson", this) {args ->
			TODO("loadVersionedJson")
		}

		state.setTableFunction("evalFunction", this) {args ->
			val name = args.getString()
			val fn = jsonFunctions[name] ?: throw NoSuchElementException("No such function $name")
			args.push(fn.value.evaluate(args.getDouble()))
			1
		}

		state.setTableFunction("evalFunction2", this) {args ->
			val name = args.getString()
			val fn = json2Functions[name] ?: throw NoSuchElementException("No such 2function $name")
			args.push(fn.value.evaluate(args.getDouble(), args.getDouble()))
			1
		}

		state.setTableFunction("imageSize", this) {args ->
			args.lua.push(imageSize(args.getString()))
			1
		}

		state.setTableFunction("imageSpaces", this) { args ->
			// List<Vec2I> root.imageSpaces(String imagePath, Vec2F worldPosition, float spaceScan, bool flip)
			val values = imageData(args.getString()).worldSpaces(args.getVector2i(), args.getDouble(), args.getBool())

			args.lua.pushTable(arraySize = values.size)
			val table = args.lua.stackTop

			for ((i, value) in values.withIndex()) {
				args.lua.push(i + 1)
				args.lua.push(value)
				args.lua.setTableValue(table)
			}

			1
		}

		state.setTableFunction("nonEmptyRegion", this) { args ->
			args.lua.push(imageData(args.getString()).nonEmptyRegion)
			1
		}

		state.setTableFunction("npcConfig", this) { args ->
			// Json root.npcConfig(String npcType)
			val name = args.getString()
			args.lua.push(npcTypes[name]?.copy() ?: throw NoSuchElementException("No such NPC type $name"))
			1
		}

		state.setTableFunction("npcVariant", this) { args ->
			// Json root.npcVariant(String species, String npcType, float level, [unsigned seed], [Json parameters])
			TODO()
		}

		state.setTableFunction("projectileGravityMultiplier", this) { args ->
			// float root.projectileGravityMultiplier(String projectileName)
			TODO()
		}

		state.setTableFunction("projectileConfig", this) { args ->
			// Json root.projectileConfig(String projectileName)
			val name = args.getString()
			args.lua.push(projectiles[name]?.copy() ?: throw kotlin.NoSuchElementException("No such Projectile type $name"))
			1
		}

		state.setTableFunction("recipesForItem", this) { args ->
			args.lua.push(JsonArray().also { a ->
				recipeRegistry.output2recipes[args.getString()]?.stream()?.map { gson.toJsonTree(it) }?.forEach {
					a.add(it)
				}
			})

			1
		}

		state.pop()

		state.load(polyfill, "@starbound.jar!/scripts/polyfill.lua")
		state.call()
	}

	private val archivePaths = ArrayList<File>()
	private val fileSystems = ArrayList<IStarboundFile>()

	fun addFilePath(path: File) {
		fileSystems.add(PhysicalFile(path))
	}

	fun addPak(pak: StarboundPak) {
		fileSystems.add(pak.root)
	}

	override fun exists(path: String): Boolean {
		@Suppress("name_shadowing")
		var path = path

		if (path[0] == '/') {
			path = path.substring(1)
		}

		for (fs in fileSystems) {
			if (fs.locate(path).exists) {
				return true
			}
		}

		return false
	}

	override fun locate(path: String): IStarboundFile {
		@Suppress("name_shadowing")
		var path = path

		if (path[0] == '/') {
			path = path.substring(1)
		}

		for (fs in fileSystems) {
			val file = fs.locate(path)

			if (file.exists) {
				return file
			}
		}

		return NonExistingFile(path.split("/").last(), fullPath = path)
	}

	fun locate(vararg path: String): IStarboundFile {
		for (p in path) {
			val get = locate(p)

			if (get.exists) {
				return get
			}
		}

		return NonExistingFile(path[0].split("/").last(), fullPath = path[0])
	}

	/**
	 * Добавляет pak к чтению при initializeGame
	 */
	fun addPakPath(pak: File) {
		archivePaths.add(pak)
	}

	fun getTileDefinition(name: String) = tiles[name]
	private val initCallbacks = ArrayList<() -> Unit>()

	var playerDefinition: PlayerDefinition by WriteOnce()
		private set

	private fun loadStage(
		callback: (Boolean, Boolean, String) -> Unit,
		loader: ((String) -> Unit) -> Unit,
		name: String,
	) {
		if (terminateLoading)
			return

		val time = System.currentTimeMillis()
		callback(false, false, "Loading $name...")
		logger.info("Loading $name...")

		loader {
			if (terminateLoading) {
				throw InterruptedException("Game is terminating")
			}

			callback(false, true, it)
		}

		callback(false, true, "Loaded $name in ${System.currentTimeMillis() - time}ms")
		logger.info("Loaded $name in ${System.currentTimeMillis() - time}ms")
	}

	private fun <T : Any> loadStage(
		callback: (Boolean, Boolean, String) -> Unit,
		registry: ObjectRegistry<T>,
		files: List<IStarboundFile>,
	) {
		loadStage(callback, loader = {
			for (listedFile in files) {
				try {
					it("Loading $listedFile")
					registry.add(gson, listedFile, pathStack)
				} catch (err: Throwable) {
					logger.error("Loading ${registry.name} definition file $listedFile", err)
				}

				if (terminateLoading) {
					break
				}
			}
		}, registry.name)
	}

	private fun doInitialize(callback: (finished: Boolean, replaceStatus: Boolean, status: String) -> Unit) {
		var time = System.currentTimeMillis()

		if (archivePaths.isNotEmpty()) {
			callback(false, false, "Searching for pak archives...".also(logger::info))

			for (path in archivePaths) {
				callback(false, false, "Reading index of ${path}...".also(logger::info))

				addPak(StarboundPak(path) { _, status ->
					callback(false, true, "${path.parent}/${path.name}: $status")
				})
			}
		}

		callback(false, false, "Finished reading pak archives in ${System.currentTimeMillis() - time}ms".also(logger::info))
		time = System.currentTimeMillis()
		callback(false, false, "Building file index...".also(logger::info))

		val ext2files = fileSystems.parallelStream()
			.flatMap { it.explore() }
			.filter { it.isFile }
			.collect(object :
					Collector<IStarboundFile, Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>,
					Map<String, List<IStarboundFile>>>,
					Supplier<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>>,
					BiConsumer<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, IStarboundFile>,
					BinaryOperator<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>>
			{
				override fun accept(t: Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, u: IStarboundFile) {
					t.computeIfAbsent(u.name.substringAfterLast('.'), Object2ObjectFunction { ArrayList() }).add(u)
				}

				override fun get(): Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>> {
					return Object2ObjectOpenHashMap()
				}

				override fun supplier(): Supplier<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>> {
					return this
				}

				override fun accumulator(): BiConsumer<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, IStarboundFile> {
					return this
				}

				override fun apply(
					t: Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>,
					u: Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>
				): Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>> {
					for ((k, v) in u)
						t.computeIfAbsent(k, Object2ObjectFunction { ArrayList() }).addAll(v)

					return t
				}

				override fun combiner(): BinaryOperator<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>> {
					return this
				}

				override fun finisher(): Function<Object2ObjectOpenHashMap<String, ArrayList<IStarboundFile>>, Map<String, List<IStarboundFile>>> {
					return Function { it }
				}

				override fun characteristics(): Set<Collector.Characteristics> {
					return setOf(Collector.Characteristics.IDENTITY_FINISH, Collector.Characteristics.UNORDERED)
				}
			})

		callback(false, false, "Finished building file index in ${System.currentTimeMillis() - time}ms".also(logger::info))

		loadStage(callback, { loadItemDefinitions(it, ext2files) }, "item definitions")
		loadStage(callback, { loadJsonFunctions(it, ext2files["functions"] ?: listOf()) }, "json functions")
		loadStage(callback, { loadJson2Functions(it, ext2files["2functions"] ?: listOf()) }, "json 2functions")
		loadStage(callback, { loadRecipes(it, ext2files["recipe"] ?: listOf()) }, "recipes")

		loadStage(callback, _tiles, ext2files["material"] ?: listOf())
		loadStage(callback, _tileModifiers, ext2files["matmod"] ?: listOf())
		loadStage(callback, _liquid, ext2files["liquid"] ?: listOf())
		loadStage(callback, _statusEffects, ext2files["statuseffect"] ?: listOf())
		loadStage(callback, _species, ext2files["species"] ?: listOf())
		loadStage(callback, _particles, ext2files["particle"] ?: listOf())
		loadStage(callback, _questTemplates, ext2files["questtemplate"] ?: listOf())
		loadStage(callback, _techs, ext2files["tech"] ?: listOf())
		loadStage(callback, _npcTypes, ext2files["npctype"] ?: listOf())
		loadStage(callback, _projectiles, ext2files["projectile"] ?: listOf())

		pathStack.block("/") {
			//playerDefinition = gson.fromJson(locate("/player.config").reader(), PlayerDefinition::class.java)
		}

		initializing = false
		initialized = true
		callback(true, false, "Finished loading in ${System.currentTimeMillis() - time}ms")
	}

	fun initializeGame(callback: (finished: Boolean, replaceStatus: Boolean, status: String) -> Unit) {
		if (initializing) {
			throw IllegalStateException("Already initializing!")
		}

		if (initialized) {
			throw IllegalStateException("Already initialized!")
		}

		initializing = true
		Thread({ doInitialize(callback) }, "Asset Loader").also {
			it.isDaemon = true
			it.start()
		}
	}

	fun onInitialize(callback: () -> Unit) {
		if (initialized) {
			callback()
		} else {
			initCallbacks.add(callback)
		}
	}

	fun pollCallbacks() {
		if (initialized && initCallbacks.isNotEmpty()) {
			for (callback in initCallbacks) {
				callback()
			}

			initCallbacks.clear()
		}
	}

	private fun loadItemDefinitions(callback: (String) -> Unit, files: Map<String, Collection<IStarboundFile>>) {
		val fileMap = mapOf(
			"item" to ItemPrototype::class.java,
			"currency" to CurrencyItemPrototype::class.java,
			"liqitem" to LiquidItemPrototype::class.java,
			"matitem" to MaterialItemPrototype::class.java,
			"flashlight" to FlashlightPrototype::class.java,
			"harvestingtool" to HarvestingToolPrototype::class.java,
			"head" to HeadArmorItemPrototype::class.java,
			"chest" to ChestArmorItemPrototype::class.java,
			"legs" to LegsArmorItemPrototype::class.java,
			"back" to BackArmorItemPrototype::class.java,
		)

		for ((ext, clazz) in fileMap) {
			val fileList = files[ext] ?: continue

			for (listedFile in fileList) {
				try {
					callback("Loading $listedFile")
					val json = gson.fromJson(listedFile.reader(), JsonObject::class.java)
					val def: ItemPrototype = pathStack(listedFile.computeDirectory()) { gson.fromJson(JsonTreeReader(json), clazz) }
					_items.add(def, json, listedFile, gson, pathStack)
				} catch (err: Throwable) {
					logger.error("Loading item definition file $listedFile", err)
				}

				if (terminateLoading) {
					return
				}
			}
		}
	}

	private fun loadJsonFunctions(callback: (String) -> Unit, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			try {
				callback("Loading $listedFile")
				val json = gson.fromJson(listedFile.reader(), JsonObject::class.java)

				for ((k, v) in json.entrySet()) {
					val fn = gson.fromJson<JsonFunction>(JsonTreeReader(v), JsonFunction::class.java)
					_jsonFunctions.add(fn, json, listedFile, gson, pathStack, k)
				}
			} catch (err: Throwable) {
				logger.error("Loading json function definition file $listedFile", err)
			}

			if (terminateLoading) {
				return
			}
		}
	}

	private fun loadJson2Functions(callback: (String) -> Unit, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			try {
				callback("Loading $listedFile")
				val json = gson.fromJson(listedFile.reader(), JsonObject::class.java)

				for ((k, v) in json.entrySet()) {
					val fn = gson.fromJson<Json2Function>(JsonTreeReader(v), Json2Function::class.java)
					_json2Functions.add(fn, json, listedFile, gson, pathStack, k)
				}
			} catch (err: Throwable) {
				logger.error("Loading json 2function definition file $listedFile", err)
			}

			if (terminateLoading) {
				return
			}
		}
	}

	private fun loadRecipes(callback: (String) -> Unit, files: Collection<IStarboundFile>) {
		for (listedFile in files) {
			try {
				callback("Loading $listedFile")
				recipeRegistry.add(gson.fromJson(listedFile.reader(), RecipeDefinition::class.java))
			} catch (err: Throwable) {
				logger.error("Loading recipe definition file $listedFile", err)
			}

			if (terminateLoading) {
				return
			}
		}
	}

	companion object {
		/**
		 * Глобальный [Interner] для [String]
		 *
		 * Так как нет смысла иметь множество [Interner]'ов для потенциального "больше одного" [Starbound],
		 * данный [Interner] доступен глобально
		 */
		@JvmField
		val STRINGS: Interner<String> = Interners.newBuilder().weak().concurrencyLevel(8).build()

		private val polyfill by lazy { loadInternalScript("polyfill") }
	}
}
